//
//  MyController.m - the core of the program
//  VMTester
//
//  Created by Cameron Hayne on 28/11/04.
//  Copyright 2006 Hayne of Tintagel.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#import "MyController.h"
#import "resourceLimits.h"

// private MyController methods
@interface MyController (Private)
- (void)updateInfo:(NSTimer *)theTimer;
- (void)updateInfoNow;
- (BOOL)allocateMemory:(float)numMegaBytes;
- (void)updateDisplay;
- (void)updateTotalMemAlloced;
- (void)setButtonsState;
- (void)setTableColour:(NSColor *) aColor;
- (void)deselectAllTableRows;
@end

@implementation MyController

- (void)awakeFromNib
{
    // configuration parameters
    myMaxHistoryItems = 10000;
    myUpdatePeriod = 5.0; // seconds
    myMemoryReadPeriod = 15.0; // seconds
    
    mySysInfoCtr = [[SysInfoCtr alloc] init];
    int pageSize = [mySysInfoCtr pageSizeMB];
    myProgInfoCtr = [[ProgInfoCtr alloc] initWithPageSizeMB: pageSize];
    myAllocInfoCtr = [[AllocInfoCtr alloc] init];
    myAllocations = [[Allocations alloc] init];
    myHistory = [[HTRingBuffer alloc] initWithCapacity: myMaxHistoryItems];
    myAutoUpdate = YES;
    myHistoryColour = [[NSColor lightGrayColor] retain];
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    printResourceLimits(); // just for interest
    
    // set the data sources & delegates for the various NSTable's
    [mySysInfoTbl setDataSource: mySysInfoCtr];
    [myProgInfoTbl setDataSource: myProgInfoCtr];
    [myMemAllocTbl setDataSource: myAllocInfoCtr];
    [myMemAllocTbl setDelegate: self];
    
    [self setButtonsState];
    
    // schedule an update every 5 seconds
    [NSTimer scheduledTimerWithTimeInterval: myUpdatePeriod
                                     target: self
                                   selector: @selector(updateInfo:)
                                   userInfo: nil
                                    repeats: YES];
    
    // schedule a read from the allocated memory every 15 seconds
    // in order to make it seem like it is being used
    [NSTimer scheduledTimerWithTimeInterval: myMemoryReadPeriod
                                     target: self
                                   selector: @selector(readFromAllocatedMemory:)
                                   userInfo: nil
                                    repeats: YES];
    
    // do an initial "update"
    [self updateInfo: nil];
}

// closing the window quits the app
- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApp
{
    return YES;
}

#pragma mark Update methods

// called from the timer that is set up in 'applicationDidFinishLaunching'
//    and from 'applicationDidFinishLaunching' directly
//    and from 'updateInfoNow'
- (void)updateInfo:(NSTimer *)aTimer
{
    Info *info = [[Info alloc] initWithAllocations: [myAllocations allocInfoArray]];
    [myHistory addObject: info];
    [info release]; // it is retained by the history list

    if (myAutoUpdate)
    {
        [myHistory setCurrentIndexToLast];
        [self updateDisplay];
    }
}

- (void)updateInfoNow
{
    [self updateInfo: nil];
}

// called from the timer that is set up in 'applicationDidFinishLaunching'
- (void)readFromAllocatedMemory:(NSTimer *)aTimer
{
    double total = [myAllocations readFromAllocatedMemory];
    //NSLog(@"total: %g", total);
    if (total == 3.1415926)
    {
        // we do this to ensure that this doesn't get optimized away
        NSLog(@"The total came out to Pi !");
    }
}

- (void)updateDisplay
{
    Info *info = [myHistory currentObject];
    Info *prevInfo = [myHistory previousObject];
    
    [myDateTimeLbl setStringValue: [info dateString]];
    
    [mySysInfoCtr setCurrentInfo: [info sysInfo]
                        PrevInfo: [prevInfo sysInfo]];
    
    [myProgInfoCtr setCurrentInfo: [info progInfo]
                         PrevInfo: [prevInfo progInfo]];
    
    [myAllocInfoCtr setCurrentInfo: [info allocations]
                          PrevInfo: [prevInfo allocations]];
    
    [mySysInfoTbl reloadData];
    [myProgInfoTbl reloadData];
    [myMemAllocTbl reloadData];
    [self updateTotalMemAlloced];
    
    [self setButtonsState];
}

- (void)updateTotalMemAlloced
{
    float total = [myAllocInfoCtr currTotalMegaBytesAlloced];
    [myTotalMemAllocLbl setStringValue: [NSString stringWithFormat:
                                                @"total %.1f MB", total]];
}

// called via a timer from 'toggleAutoUpdate'
- (void)removeAutoUpdateBtn:(NSTimer *)aTimer
{
    [myAutoUpdateBtn setTransparent: YES];
    [myAutoUpdateBtn setEnabled: NO];
}

- (void)toggleAutoUpdate
{
    if (myAutoUpdate)
    {
        // turn off auto-updating and make the "Auto Update" button active
        [myAutoUpdateBtn setTransparent: NO];
        [myAutoUpdateBtn setEnabled: YES];
        
        [self deselectAllTableRows];
        [self setTableColour: [NSColor lightGrayColor]];
        myAutoUpdate = NO;
    }
    else
    {
        // turn on auto-updating and remove the "Auto Update" button
        // Note: we remove the button via a timer so as to get the right visuals
        [NSTimer scheduledTimerWithTimeInterval: 0
                                         target: self
                                       selector: @selector(removeAutoUpdateBtn:)
                                       userInfo: nil
                                        repeats: NO];
        
        [self setTableColour: [NSColor whiteColor]];
        
        myAutoUpdate = YES;
    }
    
    [self setButtonsState];
}

- (void)setButtonsState
{    
    [myResetPageInsOutsBtn setEnabled: myAutoUpdate];
    [myAllocateBtn setEnabled: myAutoUpdate];
    [myAllocAmountTxf setEnabled: myAutoUpdate];
    
    BOOL somethingAllocated = [myAllocations numAllocs] > 0;
    [myDeAllocateAllBtn setEnabled: (myAutoUpdate && somethingAllocated)];
        
    BOOL anAllocIsSelected = ([myMemAllocTbl selectedRow] >= 0);
    [myDeAllocateSelectedBtn setEnabled: (myAutoUpdate && anAllocIsSelected)];
    
    BOOL taskIsRunning = [myConsolidationTask isRunning];
    [myConsolidateBtn setEnabled: (myAutoUpdate && !taskIsRunning)];
}

- (void)deselectAllTableRows
{
    [mySysInfoTbl  deselectAll: nil];
    [myProgInfoTbl deselectAll: nil];
    [myMemAllocTbl deselectAll: nil];
}

- (void)setTableColour:(NSColor *)aColor
{
    [mySysInfoTbl  setBackgroundColor: aColor];
    [myProgInfoTbl setBackgroundColor: aColor];
    [myMemAllocTbl setBackgroundColor: aColor];
}

#pragma mark Notifications

// this is called since we are the delegate for some of the tables
- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
    [self setButtonsState];
}

#pragma mark Actions

- (void)exportToFile:(NSString *)filename
{
    NSFileHandle *fh = [NSFileHandle fileHandleForWritingAtPath: filename];
    if (fh == nil)
    {
        NSLog(@"VMTester:exportToFile: failed to open file for writing");
        return;
    }

    NSString *legend = [NSString stringWithFormat: @"%@,%@,%@,%@\n",
                        @"date/time",
                        @"epoch seconds",
                        [mySysInfoCtr commaSeparatedLegend],
                        [myProgInfoCtr commaSeparatedLegend]
                        ];
    [fh writeData: [legend dataUsingEncoding: NSASCIIStringEncoding]];
    
    int count = [myHistory count];
    for (int i = 0; i < count; i++)
    {
        Info *info = [myHistory objectAtIndex: i];
        NSString *dateStr = [info dateString];
        NSTimeInterval epochSec = [info dateInEpochSec];

        SysInfo *sys = [info sysInfo];
        ProgInfo *prog = [info progInfo];
        NSString *str = [NSString stringWithFormat: @"%@,%f,%@,%@\n",
                        dateStr,
                        epochSec,
                        [mySysInfoCtr sysInfoCommaSeparated: sys],
                        [myProgInfoCtr progInfoCommaSeparated: prog]
                        ];
        [fh writeData: [str dataUsingEncoding: NSASCIIStringEncoding]];
    }
    
    [fh closeFile];
}

- (IBAction)turnOnAutoUpdate:(id)sender
{
    if ( ! myAutoUpdate)
    {
        [self toggleAutoUpdate];
    }
    
    [myHistory setCurrentIndexToLast];
    [self updateDisplay];
}

- (IBAction)showPreviousInfo:(id)sender
{
    if (myAutoUpdate)
    {
        // turn off auto-update
        [self toggleAutoUpdate];
    }
    
    BOOL optionKeyDown = (([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) != 0);
    if (optionKeyDown)
    {
        [myHistory setCurrentIndexBackBy: 10];
    }
    else
    {
        [myHistory setCurrentIndexToPrevious];
    }
    [self updateDisplay];
}

- (IBAction)showNextInfo:(id)sender
{
    if (myAutoUpdate)
    {
        // turn off auto-update
        [self toggleAutoUpdate];
    }

    BOOL optionKeyDown = (([[NSApp currentEvent] modifierFlags] & NSAlternateKeyMask) != 0);
    if (optionKeyDown)
    {
        [myHistory setCurrentIndexForwardBy: 10];
    }
    else
    {
        [myHistory setCurrentIndexToNext];
    }
    [self updateDisplay];
}

- (IBAction)openConsolidateSheet:(id)sender
{
    [myConsolidateCmdTxf setStringValue: @"du -sx / 2> /dev/null"];
    
    [NSApp beginSheet: myConsolidateSheet
       modalForWindow: myWindow
        modalDelegate: self
       didEndSelector: NULL
          contextInfo: nil];
}

- (IBAction)closeConsolidateSheet:(id)sender
{
    [myConsolidateSheet orderOut: nil];
    [NSApp endSheet: myConsolidateSheet];
    
    if (sender == myConsolidateOkBtn)
    {
        NSString *cmd = [myConsolidateCmdTxf stringValue];
        NSLog(@"cmd: %@", cmd);
        NSMutableArray *args = [NSMutableArray array];
        [args addObject: @"-c"];
        [args addObject: cmd];
        
        [myConsolidationTask release];
        myConsolidationTask = [NSTask launchedTaskWithLaunchPath: @"/bin/sh"
                                                       arguments: args];
        [myConsolidationTask retain];
        
        [self updateInfoNow];
    }
}

- (IBAction)resetPageInsOutsAction:(id)sender
{
    [SysInfo resetPageInsOutsBaselines];
    [self updateInfoNow];
}

- (IBAction)deallocateSelectedAction:(id)sender
{
    int row = [myMemAllocTbl selectedRow];
    if (row >= 0)
    {
        [myAllocations deallocateEntryAtIndex: row];
        [self updateInfoNow];
    }
}

- (IBAction)deallocateAllAction:(id)sender
{
    [myAllocations deallocateAll];
    [self updateInfoNow];
}

- (void)_allocate:(bool)used
{
    float numMegaBytes = [myAllocAmountTxf floatValue];
    //NSLog(@"numMegaBytes: %f", numMegaBytes);
    if (numMegaBytes <= 0.0)
    {
        return;
    }
    
    if ([myAllocations allocateMemory: numMegaBytes
                                 Used: used] == NO)
    {
        //NSLog(@"Allocation failed (numMegaBytes = %f)", numMegaBytes);
        NSBeginAlertSheet(@"Memory Allocation Failed",
                          nil, nil, nil,
                          myWindow,
                          self, nil, nil,
                          NULL,
                          @"The attempt to allocate %g MB failed",
                          numMegaBytes);
    }
    
    printResourceLimits(); // just for interest
    
    [self updateInfoNow];
}

- (IBAction)allocateAction:(id)sender
{
    [self _allocate: NO];
}

- (IBAction)allocateAndUseAction:(id)sender
{
    [self _allocate: YES];
}

- (IBAction)exportAction:(id)sender
{
    NSSavePanel *panel = [NSSavePanel savePanel];
    [panel setRequiredFileType:@"txt"];
    [panel beginSheetForDirectory: nil
                             file: @"vmtesterData.txt"
                   modalForWindow: myWindow
                    modalDelegate: self
                   didEndSelector: @selector(exportDidEnd:returnCode:contextInfo:) 
                      contextInfo: nil];
}

- (void)exportDidEnd:(NSSavePanel *)sheet returnCode:(int)code contextInfo:(void *)info
{
    if (code == NSOKButton)
    {
        NSString *filename = [sheet filename];
        NSFileManager *fm = [NSFileManager defaultManager];
        BOOL status = [fm createFileAtPath: filename
                                  contents: nil
                                attributes: nil];
        if (status)
        {
            [self exportToFile: filename];
        }
        else
        {
            NSLog(@"VMTester:exportDidEnd: failed to create file");
        }
    }
}

@end
